@@ -1,31 +1,65 @@
-import graphene
 import inspect
-from typing import Type
-from types import MethodType
 from collections.abc import Iterable
+from types import MethodType
+from typing import Any, Dict, Type, Union
 
+import graphene
+from django.apps import apps
 from django.db import models
-from django.db.models.query import QuerySet
-from django.conf import settings
-from django.contrib.contenttypes.models import ContentType
 from django.template.loader import render_to_string
-from wagtail.contrib.settings.models import BaseSetting
-from wagtail.core.models import Page as WagtailPage
-from wagtail.core.blocks import BaseBlock, RichTextBlock, stream_block
-from wagtail.core.rich_text import RichText, expand_db_html
-from wagtail.documents.models import AbstractDocument
-from wagtail.images.models import AbstractImage, AbstractRendition
-from wagtail.images.blocks import ImageChooserBlock
-from wagtail.snippets.models import get_snippet_models
 from graphene_django.types import DjangoObjectType
 
+try:
+    from wagtail.contrib.settings.models import BaseGenericSetting, BaseSiteSetting
+except ImportError:
+    # Wagtail < 4.0
+    from wagtail.contrib.settings.models import BaseSetting as BaseGenericSetting
+    from wagtail.contrib.settings.models import BaseSetting as BaseSiteSetting
 
+
+try:
+    from wagtail.blocks import StructValue, stream_block
+    from wagtail.models import Page as WagtailPage
+    from wagtail.rich_text import RichText, expand_db_html
+except ImportError:
+    from wagtail.core.blocks import StructValue, stream_block
+    from wagtail.core.models import Page as WagtailPage
+    from wagtail.core.rich_text import RichText, expand_db_html
+
+from wagtail.documents.models import AbstractDocument
+from wagtail.images.blocks import ImageChooserBlock
+from wagtail.images.models import AbstractImage, AbstractRendition
+from wagtail.snippets.models import get_snippet_models
+
+from .helpers import field_middlewares, streamfield_types
 from .registry import registry
-from .types.pages import PageInterface, Page
+from .settings import grapple_settings
 from .types.documents import DocumentObjectType
-from .types.streamfield import generate_streamfield_union
 from .types.images import ImageObjectType
-from .helpers import streamfield_types
+from .types.pages import Page, PageInterface
+from .types.streamfield import generate_streamfield_union
+
+if apps.is_installed("wagtailmedia"):
+    from wagtailmedia.models import AbstractMedia
+
+    from .types.media import MediaObjectType
+
+    has_wagtail_media = True
+
+else:
+    # TODO: find a better way to have this as an optional dependency
+    class AbstractMedia:
+        def __init__(self):
+            pass
+
+        def __add__(self, other):
+            pass
+
+        def __name__(self):
+            pass
+
+    MediaObjectType = None
+    has_wagtail_media = False
 
 
 def import_apps():
@@ -34,10 +68,15 @@ def import_apps():
     """
 
     # Register each app in the django project.
-    apps = settings.GRAPPLE_APPS.items()
-    for name, prefix in apps:
-        add_app(name, prefix)
-        registry.apps.append(name)
+    if isinstance(grapple_settings.APPS, (list, tuple)):
+        for name in grapple_settings.APPS:
+            add_app(name)
+            registry.apps.append(name)
+    else:
+        apps = grapple_settings.APPS.items()
+        for name, prefix in apps:
+            add_app(name, prefix)
+            registry.apps.append(name)
 
     # Register any 'decorated' streamfield structs.
     for streamfield_type in streamfield_types:
@@ -56,6 +95,8 @@ def import_apps():
 
         registry.streamfield_blocks[streamfield_type["cls"]] = node_type
 
+    registry.field_middlewares = field_middlewares
+
 
 def add_app(app_label: str, prefix: str = ""):
     """
@@ -67,7 +108,7 @@ def add_app(app_label: str, prefix: str = ""):
     app = apps.get_app_config(app_label)
 
     # Create a collection of models of standard models (Pages, Images, Documents).
-    models = [mdl for mdl in app.get_models()]
+    models = list(app.get_models())
 
     # Add snippet models to model collection.
     for snippet in get_snippet_models():
@@ -89,12 +130,14 @@ def register_model(cls: type, type_prefix: str):
         if issubclass(cls, WagtailPage):
             register_page_model(cls, type_prefix)
         elif issubclass(cls, AbstractDocument):
-            register_documment_model(cls, type_prefix)
+            register_document_model(cls, type_prefix)
         elif issubclass(cls, AbstractImage):
             register_image_model(cls, type_prefix)
         elif issubclass(cls, AbstractRendition):
             register_image_model(cls, type_prefix)
-        elif issubclass(cls, BaseSetting):
+        elif has_wagtail_media and issubclass(cls, AbstractMedia):
+            register_media_model(cls, type_prefix)
+        elif issubclass(cls, (BaseSiteSetting, BaseGenericSetting)):
             register_settings_model(cls, type_prefix)
         elif cls in get_snippet_models():
             register_snippet_model(cls, type_prefix)
@@ -106,9 +149,12 @@ def get_fields_and_properties(cls):
     """
     Return all fields and @property methods for a model.
     """
-    fields = [field.name for field in cls._meta.get_fields(include_parents=False)]
+    from graphene_django.utils import get_model_fields
+
+    # Note: graphene-django use this method to get the model fields
+    # cls._meta.get_fields(include_parents=False) includes symmetrical ManyToMany fields, while get_model_fields doesn't
+    fields = [field for field, instance in get_model_fields(cls)]
 
-    properties = []
     try:
         properties = [
             method[0]
@@ -195,17 +241,21 @@ def build_node_type(
 ):
     """
     Build a graphene node type from a model class and associate
-    with an interface. If it has custom fields then implmement them.
+    with an interface. If it has custom fields then implement them.
     """
     type_name = type_prefix + cls.__name__
 
-    # Create a tempory model and tempory node that will be replaced later on.
-    class StubModel(models.Model):
-        class Meta:
-            managed = False
+    # Create a temporary model and temporary node that will be replaced later on.
+    class UnmanagedMeta:
+        app_label = type_name
+        managed = False
+
+    stub_model = type(
+        type_name, (models.Model,), {"__module__": "", "Meta": UnmanagedMeta}
+    )
 
     class StubMeta:
-        model = StubModel
+        model = stub_model
 
     type_meta = {
         "Meta": StubMeta,
@@ -237,10 +287,26 @@ def load_type_fields():
                 # Recreate the graphene type with the fields set
                 class Meta:
                     model = cls
-                    interfaces = (interface,) if interface is not None else tuple()
+                    interfaces = (interface,) if interface is not None else ()
 
                 type_meta = {"Meta": Meta, "id": graphene.ID(), "name": type_name}
-                exclude_fields = get_fields_and_properties(cls)
+
+                exclude_fields = []
+                base_type_for_exclusion_checks = (
+                    base_type if not issubclass(cls, WagtailPage) else WagtailPage
+                )
+                for field in get_fields_and_properties(cls):
+                    # Filter out any fields that are defined on the interface of base type to prevent the
+                    # 'Excluding the custom field "<field>" on DjangoObjectType "<cls>" has no effect.
+                    # Either remove the custom field or remove the field from the "exclude" list.' warning
+                    if (
+                        field == "id"
+                        or hasattr(interface, field)
+                        or hasattr(base_type_for_exclusion_checks, field)
+                    ):
+                        continue
+
+                    exclude_fields.append(field)
 
                 # Add any custom fields to node if they are defined.
                 methods = {}
@@ -279,33 +345,92 @@ def convert_to_underscore(name):
     return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()
 
 
+def get_field_value(instance, field_name: str):
+    """
+    Returns the value of a given field on an object of a StreamField.
+
+    Different types of objects require different ways to access the values.
+    """
+    if isinstance(instance, StructValue):
+        return instance[field_name]
+    elif isinstance(instance.value, RichText):
+        # Allow custom markup for RichText
+        return render_to_string(
+            "wagtailcore/richtext.html", {"html": expand_db_html(instance.value.source)}
+        )
+    elif isinstance(instance.value, stream_block.StreamValue):
+        stream_data = dict(instance.value.stream_data)
+        return stream_data[field_name]
+    else:
+        return instance.value[field_name]
+
+
+def get_all_field_values(*, instance, cls) -> Dict[str, Any]:
+    """
+    Returns a dictionary of all fields and their values within a given stream
+    block instance.
+    """
+
+    values: Dict[str, Any] = {}
+    for item in cls.base_blocks.items():
+        field_name: str = item[0]
+        values[field_name] = get_field_value(
+            instance=instance,
+            field_name=field_name,
+        )
+
+    return values
+
+
 def streamfield_resolver(self, instance, info, **kwargs):
     value = None
     if hasattr(instance, "block"):
         field_name = convert_to_underscore(info.field_name)
         block = instance.block.child_blocks[field_name]
 
-        if isinstance(instance.value, stream_block.StreamValue):
-            stream_data = dict(instance.value.stream_data)
-            value = stream_data[field_name]
-        else:
-            value = instance.value[field_name]
-
-        # Allow custom markup for RichText
-        if isinstance(value, RichText):
-            value = render_to_string(
-                "wagtailcore/richtext.html", {"html": expand_db_html(value.source)}
-            )
-
-        if not block or not value:
+        if not block:
             return None
 
+        value = get_field_value(instance, field_name)
         if issubclass(type(block), ImageChooserBlock) and isinstance(value, int):
             return block.to_python(value)
 
     return value
 
 
+def custom_cls_resolver(*, cls, graphql_field):
+    klass = cls()
+
+    # If we've defined a `source` kwarg: use it.
+    if hasattr(graphql_field, "field_source") and hasattr(
+        klass, graphql_field.field_source
+    ):
+        if isinstance(getattr(type(cls()), graphql_field.field_source), property):
+            return lambda self, instance, info, **kwargs: getattr(
+                klass, graphql_field.field_source
+            )
+        elif callable(getattr(cls, graphql_field.field_source)):
+            return lambda self, instance, info, **kwargs: getattr(
+                klass, graphql_field.field_source
+            )(values=get_all_field_values(instance=instance, cls=cls))
+
+    # If the `field_name` is a property or method of the class: use it.
+    if hasattr(graphql_field, "field_name") and hasattr(
+        klass, graphql_field.field_name
+    ):
+        if isinstance(getattr(type(cls()), graphql_field.field_name), property):
+            return lambda self, instance, info, **kwargs: getattr(
+                klass, graphql_field.field_name
+            )
+        elif callable(getattr(cls, graphql_field.field_name)):
+            return lambda self, instance, info, **kwargs: getattr(
+                klass, graphql_field.field_name
+            )(values=get_all_field_values(instance=instance, cls=cls))
+
+    # No match found - fall back to the streamfield_resolver() later.
+    return None
+
+
 def build_streamfield_type(
     cls: type,
     type_prefix: str,
@@ -323,7 +448,7 @@ def build_streamfield_type(
                 registry.streamfield_blocks.get(block) for block in cls.graphql_types
             ]
         else:
-            interfaces = (interface,) if interface is not None else tuple()
+            interfaces = (interface,) if interface is not None else ()
 
     methods = {}
     type_name = type_prefix + cls.__name__
@@ -331,15 +456,17 @@ def build_streamfield_type(
 
     # Add any custom fields to node if they are defined.
     if hasattr(cls, "graphql_fields"):
-        for field in cls.graphql_fields:
-            if callable(field):
-                field = field()
+        for item in cls.graphql_fields:
+            if callable(item):
+                item = item()
 
             # Get correct types from field
-            field, field_type = get_field_type(field)
+            field, field_type = get_field_type(item)
 
             # Add support for `graphql_fields`
-            methods["resolve_" + field.field_name] = streamfield_resolver
+            methods["resolve_" + field.field_name] = (
+                custom_cls_resolver(cls=cls, graphql_field=item) or streamfield_resolver
+            )
 
             # Add field to GQL type with correct field-type
             type_meta[field.field_name] = field_type
@@ -370,7 +497,7 @@ def register_page_model(cls: Type[WagtailPage], type_prefix: str):
         registry.pages[cls] = page_node_type
 
 
-def register_documment_model(cls: Type[AbstractDocument], type_prefix: str):
+def register_document_model(cls: Type[AbstractDocument], type_prefix: str):
     """
     Create graphene node type for a model than inherits from AbstractDocument.
     Only one model will actually be generated because a default document model
@@ -427,7 +554,28 @@ def register_image_rendition_model(cls: Type[AbstractRendition], type_prefix: st
         registry.images[cls] = image_node_type
 
 
-def register_settings_model(cls: Type[BaseSetting], type_prefix: str):
+def register_media_model(cls: Type[AbstractMedia], type_prefix: str):
+    """
+    Create graphene node type for a model than inherits from AbstractDocument.
+    Only one model will actually be generated because a default document model
+    needs to be set in settings.
+    """
+
+    # Avoid gql type duplicates
+    if cls in registry.media:
+        return
+
+    # Create a GQL type derived from media model.
+    media_node_type = build_node_type(cls, type_prefix, None, MediaObjectType)
+
+    # Add media type to registry.
+    if media_node_type:
+        registry.media[cls] = media_node_type
+
+
+def register_settings_model(
+    cls: Union[Type[BaseSiteSetting], Type[BaseGenericSetting]], type_prefix: str
+):
     """
     Create a graphene node type for a settings page.
     """
