@@ -1,9 +1,7 @@
 import graphene
-
 from django.conf import settings
 from wagtail.contrib.redirects.models import Redirect
 
-from ..registry import registry
 from .pages import PageInterface
 
 
@@ -23,7 +21,7 @@ class RedirectType(graphene.ObjectType):
         if self.redirect_page is None:
             return self.link
 
-        return self.redirect_page.url_path
+        return self.redirect_page.url
 
     # Return the page that's being redirected to, if at all.
     def resolve_page(self, info, **kwargs):
@@ -36,6 +34,4 @@ class RedirectsQuery:
 
     # Return all redirects.
     def resolve_redirects(self, info, **kwargs):
-        redirects = Redirect.objects.select_related("redirect_page")
-
         return Redirect.objects.select_related("redirect_page")
